/*
 * Screen Saver - Lockup Desktop replacement for OS/2 and eComStation systems
 * Copyright (C) 2004-2005 Doodle
 *
 * Contact: doodle@dont.spam.me.scenergy.dfmk.hu
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <float.h>
#include <math.h>
#include <time.h>
#define INCL_DOS
#define INCL_WIN
#include <os2.h>

#include "ModuleSpec.h"

#ifndef M_PI
#define M_PI 3.14159
#endif

/* ------------------------------------------------------------------------*/

/*
 * The next some lines has to be changed according to the module
 * you want to create!
 */

/* Short module name: */
char *pchModuleName = "Cairo Clock";

/* Module description string: */
char *pchModuleDesc =
  "Clock and current date.\n"
  "Using the Cairo library.\n";

/* Module version number (Major.Minor format): */
int   iModuleVersionMajor = 1;
int   iModuleVersionMinor = 51;

/* Internal name generated from module name, to have
 * private window identifier for this kind of windows: */
char *pchSaverWindowClassName = "CairoClockWindowClass";

/* ------------------------------------------------------------------------*/

/* Some global variables, modified from outside: */
int bPauseDrawing = 0;
int bShutdownDrawing = 0;

/* And the real drawing loop: */
void CairoDrawLoop(HWND             hwndClientWindow,
                   int              iWidth,
                   int              iHeight,
                   cairo_surface_t *pCairoSurface,
                   cairo_t         *pCairoHandle)
{
  char achTimeString[256];
  time_t Time;
  struct tm *pTime;
  int iOldSec;
  cairo_text_extents_t CairoTextExtents;
  double xpos, ypos, xspeed, yspeed;
  double x, y, x2, y2;
  double dAlpha;
  int iUniSize;
  double dXAspect, dYAspect;
  int i;

  // Initialize position and speed
  srand(clock());
  xpos = 0;
  ypos = 0;
  xspeed = 0.001 * ((rand()/(RAND_MAX/2))?1:-1);
  yspeed = 0.001 * ((rand()/(RAND_MAX/2))?1:-1);

  // Do the main drawing loop as long as needed!
  while (!bShutdownDrawing)
  {
    if (bPauseDrawing)
    {
      // Do not draw anything if we're paused!
      DosSleep(250);
    }
    else
    {
      // Otherwise draw something!

      // Get current time
      Time = time(NULL);
      pTime = localtime(&Time);

      // Draw the clock!

      // Save cairo canvas state
      cairo_save(pCairoHandle);

      // Calculate screen aspect ratio
      if (iWidth<iHeight)
      {
        iUniSize = iWidth;
        dYAspect = iHeight;
        dYAspect /= iWidth;
        dXAspect = 1;
      }
      else
      {
        iUniSize = iHeight;
        dXAspect = iWidth;
        dXAspect /= iHeight;
        dYAspect = 1;
      }

      // Scale canvas so we'll have a
      // normalized coordinate system of (0;0) -> (1;1)
      cairo_scale(pCairoHandle, iUniSize, iUniSize);

      // Modify clock position, and
      xpos+=xspeed;
      ypos+=yspeed;

      // Make sure we'll bounce back sometimes.
      if ((xpos<-0.05*dXAspect) || (xpos>0.05*dXAspect))
        xspeed = -xspeed;
      if ((ypos<-0.05*dYAspect) || (ypos>0.05*dYAspect))
        yspeed = -yspeed;

      // Clear background with black
      cairo_set_source_rgb (pCairoHandle, 0, 0 ,0);
      cairo_rectangle (pCairoHandle,
                       0, 0,
                       dXAspect, dYAspect);
      cairo_fill(pCairoHandle);

      // Draw markers
      for (i=0; i<60; i++)
      {
        dAlpha = 360.0 * i / 60.0;

        if (i % 5==0)
        {
          cairo_set_line_width(pCairoHandle, 0.018);
          x = 0.32 * sin(dAlpha*M_PI/180.0);
          y = 0.32 * cos(dAlpha*M_PI/180.0);
        } else
        {
          cairo_set_line_width(pCairoHandle, 0.005);
          x = 0.37 * sin(dAlpha*M_PI/180.0);
          y = 0.37 * cos(dAlpha*M_PI/180.0);
        }
        x2 = 0.41 * sin(dAlpha*M_PI/180.0);
        y2 = 0.41 * cos(dAlpha*M_PI/180.0);

        cairo_set_source_rgb (pCairoHandle, 0.8, 0.8, 0.8);
        cairo_move_to (pCairoHandle, 0.5 * dXAspect + x + xpos, 0.5 * dYAspect - y + ypos);
        cairo_line_to(pCairoHandle, 0.5*dXAspect + x2 + xpos, 0.5 * dYAspect - y2 + ypos);
        cairo_close_path (pCairoHandle);
        cairo_stroke (pCairoHandle);
      }

      // Draw hour marker
      dAlpha = 360.0 * (pTime->tm_hour%12) / 12.0;
      dAlpha+= (pTime->tm_min%60)*0.5 + (pTime->tm_sec%60)*0.0083;
      x = 0.25 * sin(dAlpha*M_PI/180.0);
      y = 0.25 * cos(dAlpha*M_PI/180.0);
      cairo_set_line_width(pCairoHandle, 0.015);
      cairo_set_source_rgb (pCairoHandle, 1, 1, 1);
      cairo_arc(pCairoHandle, 0.5 * dXAspect + xpos, 0.5 * dYAspect + ypos, 0.020, 0, 2*M_PI);
      cairo_fill(pCairoHandle);
      cairo_move_to (pCairoHandle, 0.5 * dXAspect + xpos, 0.5 * dYAspect + ypos);
      cairo_line_to(pCairoHandle, 0.5*dXAspect + x + xpos, 0.5 * dYAspect - y + ypos);
      cairo_close_path (pCairoHandle);
      cairo_stroke (pCairoHandle);

      // Draw minute marker
      dAlpha = 360.0 * pTime->tm_min / 60.0;
      dAlpha+= (pTime->tm_sec%60)/10.0;
      x = 0.35 * sin(dAlpha*M_PI/180.0);
      y = 0.35 * cos(dAlpha*M_PI/180.0);
      cairo_set_line_width(pCairoHandle, 0.015);
      cairo_set_source_rgb (pCairoHandle, 1, 1, 1);
      cairo_arc(pCairoHandle, 0.5 * dXAspect + xpos, 0.5 * dYAspect + ypos, 0.017, 0, 2*M_PI);
      cairo_fill(pCairoHandle);
      cairo_move_to (pCairoHandle, 0.5 * dXAspect + xpos, 0.5 * dYAspect + ypos);
      cairo_line_to(pCairoHandle, 0.5*dXAspect + x + xpos, 0.5 * dYAspect - y + ypos);
      cairo_close_path (pCairoHandle);
      cairo_stroke (pCairoHandle);

      // Draw second marker
      dAlpha = 360.0 * pTime->tm_sec / 60.0;
      x = 0.35 * sin(dAlpha*M_PI/180.0);
      y = 0.35 * cos(dAlpha*M_PI/180.0);
      cairo_set_line_width(pCairoHandle, 0.01);
      cairo_set_source_rgb (pCairoHandle, 0.8, 0.1, 0.1);
      cairo_arc(pCairoHandle, 0.5 * dXAspect + xpos, 0.5 * dYAspect + ypos, 0.014, 0, 2*M_PI);
      cairo_fill(pCairoHandle);
      cairo_move_to (pCairoHandle, 0.5 * dXAspect + xpos, 0.5 * dYAspect + ypos);
      cairo_line_to(pCairoHandle, 0.5*dXAspect + x + xpos, 0.5 * dYAspect - y + ypos);
      cairo_close_path (pCairoHandle);

      cairo_stroke (pCairoHandle);

      // Now print there the time!
      strftime(achTimeString,
               sizeof(achTimeString),
               "%T",
               pTime);
      cairo_select_font_face(pCairoHandle, "Sans",
                             CAIRO_FONT_SLANT_NORMAL,
                             CAIRO_FONT_WEIGHT_BOLD);
      cairo_set_font_size(pCairoHandle,
                          0.08);

      // Get how bit the text will be!
      cairo_text_extents(pCairoHandle,
                         "00:00:00",
                         &CairoTextExtents);
                          
      // Print the text, centered!
      cairo_move_to(pCairoHandle,
                    0.5 * dXAspect - CairoTextExtents.width/2 + xpos,
                    0.5 * dYAspect + CairoTextExtents.height/2 - 0.15 + ypos);
      cairo_text_path(pCairoHandle, achTimeString);

      cairo_set_source_rgba(pCairoHandle,
                            0.2, 0.2, 0.8, 0.55);
      cairo_fill_preserve(pCairoHandle);
      cairo_set_source_rgba(pCairoHandle,
                            0.6, 0.6, 1.0, 0.35);
      cairo_set_line_width(pCairoHandle, 0.006);
      cairo_close_path(pCairoHandle);
      cairo_stroke (pCairoHandle);
      
      // Now print there the date!
      strftime(achTimeString,
               sizeof(achTimeString),
               "%x",
               pTime);
      cairo_select_font_face(pCairoHandle, "Sans",
                             CAIRO_FONT_SLANT_NORMAL,
                             CAIRO_FONT_WEIGHT_BOLD);
      cairo_set_font_size(pCairoHandle,
                          0.05);

      // Get how bit the text will be!
      cairo_text_extents(pCairoHandle,
                         achTimeString,
                         &CairoTextExtents);

      // Print the text, centered!
      cairo_move_to(pCairoHandle,
                    0.5 * dXAspect - CairoTextExtents.width/2 + xpos,
                    0.5 * dYAspect + CairoTextExtents.height/2 + 0.15 + ypos);
      cairo_text_path(pCairoHandle, achTimeString);

      cairo_set_source_rgba(pCairoHandle,
                            0.90, 0.90, 0.2, 0.75);
      cairo_fill_preserve(pCairoHandle);
      cairo_set_source_rgba(pCairoHandle,
                            0, 0, 0, 0.4);
      cairo_set_line_width(pCairoHandle, 0.002);
      cairo_close_path(pCairoHandle);
      cairo_stroke (pCairoHandle);

      // Restore canvas state to original one
      cairo_restore(pCairoHandle);

      // Make changes visible
      WinInvalidateRect(hwndClientWindow, NULL, FALSE);

    }
    // Wait for next second to come, or
    // to get an event of shutdown!
    iOldSec = pTime->tm_sec;
    do {
      Time = time(NULL);
      pTime = localtime(&Time);
      DosSleep(100);
    } while ((pTime->tm_sec == iOldSec) && (!bShutdownDrawing));
  }
  // Got shutdown event, so stopped drawing. That's all.
}

/* ------------------------------------------------------------------------*/
/* End of file */
